@@ -3,43 +3,39 @@
  *
  * This file is part of osm2pgsql (https://osm2pgsql.org/).
  *
- * Copyright (C) 2006-2021 by the osm2pgsql developer community.
+ * Copyright (C) 2006-2025 by the osm2pgsql developer community.
  * For a full list of authors see the git log.
  */
 
 #include <algorithm>
+#include <cstdint>
 #include <cstdio>
-#include <cstring>
-#include <stdexcept>
+#include <limits>
 #include <string>
-#include <utility>
 #include <vector>
 
 #include "format.hpp"
 #include "logging.hpp"
 #include "options.hpp"
+#include "pgsql-capabilities.hpp"
 #include "pgsql-helper.hpp"
 #include "table.hpp"
 #include "taginfo.hpp"
 #include "util.hpp"
-#include "wkb.hpp"
 
-table_t::table_t(std::string const &name, std::string const &type,
-                 columns_t const &columns, hstores_t const &hstore_columns,
-                 int const srid, bool const append, hstore_column hstore_mode,
+table_t::table_t(std::string const &name, std::string type, columns_t columns,
+                 hstores_t hstore_columns, int const srid, bool const append,
+                 hstore_column hstore_mode,
                  std::shared_ptr<db_copy_thread_t> const &copy_thread,
                  std::string const &schema)
-: m_target(std::make_shared<db_target_descr_t>(name.c_str(), "osm_id")),
-  m_type(type), m_srid(fmt::to_string(srid)), m_append(append),
-  m_hstore_mode(hstore_mode), m_columns(columns),
-  m_hstore_columns(hstore_columns), m_copy(copy_thread)
+: m_target(std::make_shared<db_target_descr_t>(schema, name, "osm_id")),
+  m_type(std::move(type)), m_srid(fmt::to_string(srid)), m_append(append),
+  m_hstore_mode(hstore_mode), m_columns(std::move(columns)),
+  m_hstore_columns(std::move(hstore_columns)), m_copy(copy_thread)
 {
-    m_target->schema = schema;
-
     // if we dont have any columns
     if (m_columns.empty() && m_hstore_mode != hstore_column::all) {
-        throw std::runtime_error{
-            "No columns provided for table {}."_format(name)};
+        throw fmt_error("No columns provided for table {}.", name);
     }
 
     generate_copy_column_list();
@@ -47,8 +43,8 @@ table_t::table_t(std::string const &name, std::string const &type,
 
 table_t::table_t(table_t const &other,
                  std::shared_ptr<db_copy_thread_t> const &copy_thread)
-: m_conninfo(other.m_conninfo), m_target(other.m_target), m_type(other.m_type),
-  m_srid(other.m_srid), m_append(other.m_append),
+: m_connection_params(other.m_connection_params), m_target(other.m_target),
+  m_type(other.m_type), m_srid(other.m_srid), m_append(other.m_append),
   m_hstore_mode(other.m_hstore_mode), m_columns(other.m_columns),
   m_hstore_columns(other.m_hstore_columns), m_table_space(other.m_table_space),
   m_copy(copy_thread)
@@ -56,64 +52,65 @@ table_t::table_t(table_t const &other,
     // if the other table has already started, then we want to execute
     // the same stuff to get into the same state. but if it hasn't, then
     // this would be premature.
-    if (other.m_sql_conn) {
+    if (other.m_db_connection) {
         connect();
         prepare();
     }
 }
 
-void table_t::teardown() { m_sql_conn.reset(); }
+void table_t::teardown() { m_db_connection.reset(); }
 
 void table_t::sync() { m_copy.sync(); }
 
 void table_t::connect()
 {
-    m_sql_conn.reset(new pg_conn_t{m_conninfo});
-    //let commits happen faster by delaying when they actually occur
-    m_sql_conn->exec("SET synchronous_commit = off");
+    m_db_connection =
+        std::make_unique<pg_conn_t>(m_connection_params, "out.pgsql");
 }
 
-void table_t::start(std::string const &conninfo, std::string const &table_space)
+void table_t::start(connection_params_t const &connection_params,
+                    std::string const &table_space)
 {
-    if (m_sql_conn) {
-        throw std::runtime_error{m_target->name +
-                                 " cannot start, its already started."};
+    if (m_db_connection) {
+        throw fmt_error("{} cannot start, its already started.",
+                        m_target->name());
     }
 
-    m_conninfo = conninfo;
+    m_connection_params = connection_params;
     m_table_space = tablespace_clause(table_space);
 
     connect();
-    log_info("Setting up table '{}'", m_target->name);
-    m_sql_conn->exec("SET client_min_messages = WARNING");
-    auto const qual_name = qualified_name(m_target->schema, m_target->name);
-    auto const qual_tmp_name = qualified_name(
-        m_target->schema, m_target->name + "_tmp");
+    log_info("Setting up table '{}'", m_target->name());
 
     // we are making a new table
     if (!m_append) {
-        m_sql_conn->exec(
-            "DROP TABLE IF EXISTS {} CASCADE"_format(qual_name));
+        drop_table_if_exists(*m_db_connection, m_target->schema(),
+                             m_target->name());
     }
 
     // These _tmp tables can be left behind if we run out of disk space.
-    m_sql_conn->exec("DROP TABLE IF EXISTS {}"_format(qual_tmp_name));
-    m_sql_conn->exec("RESET client_min_messages");
+    drop_table_if_exists(*m_db_connection, m_target->schema(),
+                         m_target->name() + "_tmp");
 
     //making a new table
     if (!m_append) {
         //define the new table
+        auto const qual_name =
+            qualified_name(m_target->schema(), m_target->name());
         auto sql =
-            "CREATE UNLOGGED TABLE {} (osm_id int8,"_format(qual_name);
+            fmt::format("CREATE UNLOGGED TABLE {} (osm_id int8,", qual_name);
 
         //first with the regular columns
         for (auto const &column : m_columns) {
-            sql += "\"{}\" {},"_format(column.name, column.type_name);
+            check_identifier(column.name, "column names");
+            check_identifier(column.type_name, "column types");
+            sql += fmt::format(R"("{}" {},)", column.name, column.type_name);
         }
 
         //then with the hstore columns
         for (auto const &hcolumn : m_hstore_columns) {
-            sql += "\"{}\" hstore,"_format(hcolumn);
+            check_identifier(hcolumn, "column names");
+            sql += fmt::format(R"("{}" hstore,)", hcolumn);
         }
 
         //add tags column
@@ -121,7 +118,7 @@ void table_t::start(std::string const &conninfo, std::string const &table_space)
             sql += "\"tags\" hstore,";
         }
 
-        sql += "way geometry({},{}) )"_format(m_type, m_srid);
+        sql += fmt::format("way geometry({},{}) )", m_type, m_srid);
 
         // The final tables are created with CREATE TABLE AS ... SELECT * FROM ...
         // This means that they won't get this autovacuum setting, so it doesn't
@@ -131,7 +128,12 @@ void table_t::start(std::string const &conninfo, std::string const &table_space)
         sql += m_table_space;
 
         //create the table
-        m_sql_conn->exec(sql);
+        m_db_connection->exec(sql);
+
+        if (m_srid != "4326") {
+            create_geom_check_trigger(*m_db_connection, m_target->schema(),
+                                      m_target->name(), "ST_IsValid(NEW.way)");
+        }
     }
 
     prepare();
@@ -140,36 +142,36 @@ void table_t::start(std::string const &conninfo, std::string const &table_space)
 void table_t::prepare()
 {
     //let postgres cache this query as it will presumably happen a lot
-    auto const qual_name = qualified_name(m_target->schema, m_target->name);
-    m_sql_conn->exec(
-        "PREPARE get_wkb(int8) AS SELECT way FROM {} WHERE osm_id = $1"_format(
-            qual_name));
+    auto const qual_name = qualified_name(m_target->schema(), m_target->name());
+    m_db_connection->prepare(
+        "get_wkb", "SELECT way FROM {} WHERE osm_id = $1::int8", qual_name);
 }
 
 void table_t::generate_copy_column_list()
 {
-    m_target->rows = "osm_id,";
-    //first with the regular columns
+    util::string_joiner_t joiner{',', '"'};
+
+    joiner.add("osm_id");
+
+    // first with the regular columns
     for (auto const &column : m_columns) {
-        m_target->rows += '"';
-        m_target->rows += column.name;
-        m_target->rows += "\",";
+        joiner.add(column.name);
     }
 
-    //then with the hstore columns
+    // then with the hstore columns
     for (auto const &hcolumn : m_hstore_columns) {
-        m_target->rows += '"';
-        m_target->rows += hcolumn;
-        m_target->rows += "\",";
+        joiner.add(hcolumn);
     }
 
-    //add tags column and geom column
+    // add tags column
     if (m_hstore_mode != hstore_column::none) {
-        m_target->rows += "tags,way";
-        //or just the geom column
-    } else {
-        m_target->rows += "way";
+        joiner.add("tags");
     }
+
+    // add geom column
+    joiner.add("way");
+
+    m_target->set_rows(joiner());
 }
 
 void table_t::stop(bool updateable, bool enable_hstore_index,
@@ -178,96 +180,67 @@ void table_t::stop(bool updateable, bool enable_hstore_index,
     // make sure that all data is written to the DB before continuing
     m_copy.sync();
 
-    auto const qual_name = qualified_name(m_target->schema, m_target->name);
-    auto const qual_tmp_name = qualified_name(
-        m_target->schema, m_target->name + "_tmp");
+    auto const qual_name = qualified_name(m_target->schema(), m_target->name());
+    auto const qual_tmp_name =
+        qualified_name(m_target->schema(), m_target->name() + "_tmp");
 
     if (!m_append) {
-        util::timer_t timer;
-
-        log_info("Clustering table '{}' by geometry...", m_target->name);
-
-        // Notices about invalid geometries are expected and can be ignored
-        // because they say nothing about the validity of the geometry in OSM.
-        m_sql_conn->exec("SET client_min_messages = WARNING");
-
-        std::string sql =
-            "CREATE TABLE {} {} AS SELECT * FROM {}"_format(
-                qual_tmp_name, m_table_space, qual_name);
-
         if (m_srid != "4326") {
-            // libosmium assures validity of geometries in 4326.
-            // Transformation to another projection could make the geometry
-            // invalid. Therefore add a filter to drop those.
-            sql += " WHERE ST_IsValid(way)";
+            drop_geom_check_trigger(*m_db_connection, m_target->schema(),
+                                    m_target->name());
         }
 
-        auto const postgis_version = get_postgis_version(*m_sql_conn);
+        log_info("Clustering table '{}' by geometry...", m_target->name());
 
-        sql += " ORDER BY ";
-        if (postgis_version.major == 2 && postgis_version.minor < 4) {
-            log_debug("Using GeoHash for clustering table '{}'",
-                      m_target->name);
-            if (m_srid == "4326") {
-                sql += "ST_GeoHash(way,10)";
-            } else {
-                sql += "ST_GeoHash(ST_Transform(ST_Envelope(way),4326),10)";
-            }
-            sql += " COLLATE \"C\"";
-        } else {
-            log_debug("Using native order for clustering table '{}'",
-                      m_target->name);
-            // Since Postgis 2.4 the order function for geometries gives
-            // useful results.
-            sql += "way";
-        }
+        std::string const sql =
+            fmt::format("CREATE TABLE {} {} AS SELECT * FROM {} ORDER BY way",
+                        qual_tmp_name, m_table_space, qual_name);
 
-        m_sql_conn->exec(sql);
+        m_db_connection->exec(sql);
 
-        m_sql_conn->exec("DROP TABLE {}"_format(qual_name));
-        m_sql_conn->exec("ALTER TABLE {} RENAME TO \"{}\""_format(
-            qual_tmp_name, m_target->name));
+        m_db_connection->exec("DROP TABLE {}", qual_name);
+        m_db_connection->exec(R"(ALTER TABLE {} RENAME TO "{}")", qual_tmp_name,
+                              m_target->name());
 
-        log_info("Creating geometry index on table '{}'...", m_target->name);
+        log_info("Creating geometry index on table '{}'...", m_target->name());
 
         // Use fillfactor 100 for un-updatable imports
-        m_sql_conn->exec("CREATE INDEX ON {} USING GIST (way) {} {}"_format(
-            qual_name, (updateable ? "" : "WITH (fillfactor = 100)"),
-            tablespace_clause(table_space_index)));
+        m_db_connection->exec("CREATE INDEX ON {} USING GIST (way) {} {}",
+                              qual_name,
+                              (updateable ? "" : "WITH (fillfactor = 100)"),
+                              tablespace_clause(table_space_index));
 
         /* slim mode needs this to be able to apply diffs */
         if (updateable) {
-            log_info("Creating osm_id index on table '{}'...", m_target->name);
-            m_sql_conn->exec(
-                "CREATE INDEX ON {} USING BTREE (osm_id) {}"_format(
-                    qual_name, tablespace_clause(table_space_index)));
+            log_info("Creating osm_id index on table '{}'...",
+                     m_target->name());
+            m_db_connection->exec("CREATE INDEX ON {} USING BTREE (osm_id) {}",
+                                  qual_name,
+                                  tablespace_clause(table_space_index));
             if (m_srid != "4326") {
-                create_geom_check_trigger(m_sql_conn.get(), m_target->schema,
-                                          m_target->name, "way");
+                create_geom_check_trigger(*m_db_connection, m_target->schema(),
+                                          m_target->name(),
+                                          "ST_IsValid(NEW.way)");
             }
         }
 
         /* Create hstore index if selected */
         if (enable_hstore_index) {
             log_info("Creating hstore indexes on table '{}'...",
-                     m_target->name);
+                     m_target->name());
             if (m_hstore_mode != hstore_column::none) {
-                m_sql_conn->exec(
-                    "CREATE INDEX ON {} USING GIN (tags) {}"_format(
-                        qual_name, tablespace_clause(table_space_index)));
+                m_db_connection->exec("CREATE INDEX ON {} USING GIN (tags) {}",
+                                      qual_name,
+                                      tablespace_clause(table_space_index));
             }
             for (auto const &hcolumn : m_hstore_columns) {
-                m_sql_conn->exec(
-                    "CREATE INDEX ON {} USING GIN (\"{}\") {}"_format(
-                        qual_name, hcolumn,
-                        tablespace_clause(table_space_index)));
+                m_db_connection->exec(
+                    R"(CREATE INDEX ON {} USING GIN ("{}") {})", qual_name,
+                    hcolumn, tablespace_clause(table_space_index));
             }
         }
-        log_info("Analyzing table '{}'...", m_target->name);
-        analyze_table(*m_sql_conn, m_target->schema, m_target->name);
-
-        log_info("All postprocessing on table '{}' done in {}.", m_target->name,
-                 util::human_readable_duration(timer.stop()));
+        log_info("Analyzing table '{}'...", m_target->name());
+        analyze_table(*m_db_connection, m_target->schema(), m_target->name());
     }
     teardown();
 }
@@ -375,13 +348,24 @@ void table_t::write_hstore_columns(taglist_t const &tags)
     }
 }
 
+void table_t::task_wait()
+{
+    auto const run_time = m_task_result.wait();
+    log_info("All postprocessing on table '{}' done in {}.", m_target->name(),
+             util::human_readable_duration(run_time));
+}
+
+// NOLINTBEGIN(google-runtime-int,cert-err34-c)
+// This is legacy code which will be removed anyway.
+
 /* Escape data appropriate to the type */
-void table_t::escape_type(std::string const &value, ColumnType flags)
+void table_t::escape_type(std::string const &value, column_type_t flags)
 {
     switch (flags) {
-    case ColumnType::INT: {
+    case column_type_t::INT: {
         // For integers we take the first number, or the average if it's a-b
-        long long from, to;
+        long long from = 0;
+        long long to = 0;
         // limit number of digits parsed to avoid undefined behaviour in sscanf
         int const items =
             std::sscanf(value.c_str(), "%18lld-%18lld", &from, &to);
@@ -403,7 +387,7 @@ void table_t::escape_type(std::string const &value, ColumnType flags)
         }
         break;
     }
-    case ColumnType::REAL:
+    case column_type_t::REAL:
         /* try to "repair" real values as follows:
          * assume "," to be a decimal mark which need to be replaced by "."
          * like int4 take the first number, or the average if it's a-b
@@ -415,7 +399,8 @@ void table_t::escape_type(std::string const &value, ColumnType flags)
             std::string escaped{value};
             std::replace(escaped.begin(), escaped.end(), ',', '.');
 
-            double from, to;
+            double from = NAN;
+            double to = NAN;
             int const items =
                 std::sscanf(escaped.c_str(), "%lf-%lf", &from, &to);
             if (items == 1) {
@@ -436,14 +421,14 @@ void table_t::escape_type(std::string const &value, ColumnType flags)
             }
             break;
         }
-    case ColumnType::TEXT:
+    case column_type_t::TEXT:
         m_copy.add_column(value);
         break;
     }
 }
+// NOLINTEND(google-runtime-int,cert-err34-c)
 
-table_t::wkb_reader table_t::get_wkb_reader(osmid_t id)
+pg_result_t table_t::get_wkb(osmid_t id)
 {
-    auto res = m_sql_conn->exec_prepared("get_wkb", id);
-    return wkb_reader{std::move(res)};
+    return m_db_connection->exec_prepared_as_binary("get_wkb", id);
 }
