@@ -3,7 +3,7 @@
  *
  * This file is part of osm2pgsql (https://osm2pgsql.org/).
  *
- * Copyright (C) 2006-2021 by the osm2pgsql developer community.
+ * Copyright (C) 2006-2025 by the osm2pgsql developer community.
  * For a full list of authors see the git log.
  */
 
@@ -12,13 +12,19 @@
 #include "common-import.hpp"
 #include "common-options.hpp"
 
-static testing::db::import_t db;
+#include <array>
 
-static char const *const conf_file = "test_output_flex_types.lua";
+namespace {
+
+testing::db::import_t db;
+
+char const *const CONF_FILE = "test_output_flex_types.lua";
+
+} // anonymous namespace
 
 TEST_CASE("type nil")
 {
-    testing::opt_t const options = testing::opt_t().flex(conf_file);
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
 
     REQUIRE_NOTHROW(
         db.run_import(options, "n10 v1 dV x10.0 y10.0 Ttype=nil\n"));
@@ -26,17 +32,17 @@ TEST_CASE("type nil")
     auto conn = db.db().connect();
 
     CHECK(1 == conn.get_count("nodes"));
-    CHECK(1 ==
-          conn.get_count(
-              "nodes",
-              "ttext IS NULL AND tbool IS NULL AND tint2 IS NULL "
-              "AND tint4 IS NULL AND tint8 IS NULL AND treal IS "
-              "NULL AND thstr IS NULL AND tdirn IS NULL AND tsqlt IS NULL"));
+    CHECK(1 == conn.get_count(
+                   "nodes",
+                   "ttext IS NULL AND tbool IS NULL AND tint2 IS NULL AND "
+                   "tint4 IS NULL AND tint8 IS NULL AND treal IS NULL AND "
+                   "thstr IS NULL AND tjson IS NULL AND tdirn IS NULL AND "
+                   "tsqlt IS NULL"));
 }
 
 TEST_CASE("type boolean")
 {
-    testing::opt_t const options = testing::opt_t().flex(conf_file);
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
 
     REQUIRE_NOTHROW(
         db.run_import(options, "n10 v1 dV x10.0 y10.0 Ttype=boolean\n"));
@@ -44,15 +50,17 @@ TEST_CASE("type boolean")
     auto conn = db.db().connect();
 
     CHECK(2 == conn.get_count("nodes"));
-    CHECK(1 == conn.get_count("nodes", "tbool = true AND tint2 = 1 AND tint4 = "
-                                       "1 AND tint8 = 1 AND tdirn = 1"));
-    CHECK(1 == conn.get_count("nodes", "tbool = false AND tint2 = 0 AND tint4 "
-                                       "= 0 AND tint8 = 0 AND tdirn = 0"));
+    CHECK(1 == conn.get_count("nodes", "tbool = true AND tint2 = 1 AND "
+                                       "tint4 = 1 AND tint8 = 1 AND "
+                                       "tjson = 'true'::jsonb AND tdirn = 1"));
+    CHECK(1 == conn.get_count("nodes", "tbool = false AND tint2 = 0 AND "
+                                       "tint4 = 0 AND tint8 = 0 AND "
+                                       "tjson = 'false'::jsonb AND tdirn = 0"));
 }
 
 TEST_CASE("type boolean in column where it doesn't belong")
 {
-    testing::opt_t const options = testing::opt_t().flex(conf_file);
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
 
     REQUIRE_THROWS(db.run_import(
         options, "n10 v1 dV x10.0 y10.0 Ttype=boolean-fail column=ttext\n"));
@@ -70,7 +78,7 @@ TEST_CASE("type boolean in column where it doesn't belong")
 
 TEST_CASE("type number")
 {
-    testing::opt_t const options = testing::opt_t().flex(conf_file);
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
 
     REQUIRE_NOTHROW(
         db.run_import(options, "n10 v1 dV x10.0 y10.0 Ttype=number\n"));
@@ -80,31 +88,31 @@ TEST_CASE("type number")
     CHECK(19 == conn.get_count("nodes"));
 
     // clang-format off
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-2147483649' AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 IS NULL       AND tint8 = -2147483649               AND tdirn = -1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-2147483648' AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 = -2147483648 AND tint8 = -2147483648               AND tdirn = -1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-2147483647' AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 = -2147483647 AND tint8 = -2147483647               AND tdirn = -1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-32769'      AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 = -32769      AND tint8 = -32769 AND treal = -32769 AND tdirn = -1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-32768'      AND ttext = tsqlt AND tbool = true  AND tint2 = -32768 AND tint4 = -32768      AND tint8 = -32768 AND treal = -32768 AND tdirn = -1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-32767'      AND ttext = tsqlt AND tbool = true  AND tint2 = -32767 AND tint4 = -32767      AND tint8 = -32767 AND treal = -32767 AND tdirn = -1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-2'          AND ttext = tsqlt AND tbool = true  AND tint2 = -2     AND tint4 = -2          AND tint8 = -2     AND treal =   -2   AND tdirn = -1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-1'          AND ttext = tsqlt AND tbool = true  AND tint2 = -1     AND tint4 = -1          AND tint8 = -1     AND treal =   -1   AND tdirn = -1"));
-    CHECK(1 == conn.get_count("nodes", "           tsqlt          = '-0.5'        AND ttext = tsqlt AND tbool = true  AND tint2 =  0     AND tint4 =  0          AND tint8 =  0     AND treal = -0.5   AND tdirn = -1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '0'           AND ttext = tsqlt AND tbool = false AND tint2 =  0     AND tint4 =  0          AND tint8 =  0     AND treal =    0   AND tdirn = 0"));
-    CHECK(1 == conn.get_count("nodes", "           tsqlt          = '0.5'         AND ttext = tsqlt AND tbool = true  AND tint2 =  0     AND tint4 =  0          AND tint8 =  0     AND treal =  0.5   AND tdirn = 1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '1'           AND ttext = tsqlt AND tbool = true  AND tint2 =  1     AND tint4 =  1          AND tint8 =  1     AND treal =    1   AND tdirn = 1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '2'           AND ttext = tsqlt AND tbool = true  AND tint2 =  2     AND tint4 =  2          AND tint8 =  2     AND treal =    2   AND tdirn = 1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '32767'       AND ttext = tsqlt AND tbool = true  AND tint2 = 32767  AND tint4 = 32767       AND tint8 = 32767  AND treal = 32767  AND tdirn = 1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '32768'       AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 = 32768       AND tint8 = 32768  AND treal = 32768  AND tdirn = 1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '32769'       AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 = 32769       AND tint8 = 32769  AND treal = 32769  AND tdirn = 1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '2147483647'  AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 = 2147483647  AND tint8 = 2147483647                AND tdirn = 1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '2147483648'  AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 IS NULL       AND tint8 = 2147483648                AND tdirn = 1"));
-    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '2147483649'  AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 IS NULL       AND tint8 = 2147483649                AND tdirn = 1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-2147483649' AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 IS NULL       AND tint8 = -2147483649               AND tjson = '-2147483649'::jsonb AND tdirn = -1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-2147483648' AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 = -2147483648 AND tint8 = -2147483648               AND tjson = '-2147483648'::jsonb AND tdirn = -1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-2147483647' AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 = -2147483647 AND tint8 = -2147483647               AND tjson = '-2147483647'::jsonb AND tdirn = -1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-32769'      AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 = -32769      AND tint8 = -32769 AND treal = -32769 AND tjson =      '-32769'::jsonb AND tdirn = -1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-32768'      AND ttext = tsqlt AND tbool = true  AND tint2 = -32768 AND tint4 = -32768      AND tint8 = -32768 AND treal = -32768 AND tjson =      '-32768'::jsonb AND tdirn = -1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-32767'      AND ttext = tsqlt AND tbool = true  AND tint2 = -32767 AND tint4 = -32767      AND tint8 = -32767 AND treal = -32767 AND tjson =      '-32767'::jsonb AND tdirn = -1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-2'          AND ttext = tsqlt AND tbool = true  AND tint2 = -2     AND tint4 = -2          AND tint8 = -2     AND treal =   -2   AND tjson =          '-2'::jsonb AND tdirn = -1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '-1'          AND ttext = tsqlt AND tbool = true  AND tint2 = -1     AND tint4 = -1          AND tint8 = -1     AND treal =   -1   AND tjson =          '-1'::jsonb AND tdirn = -1"));
+    CHECK(1 == conn.get_count("nodes", "           tsqlt          = '-0.5'        AND ttext = tsqlt AND tbool = true  AND tint2 =  0     AND tint4 =  0          AND tint8 =  0     AND treal = -0.5   AND tjson =        '-0.5'::jsonb AND tdirn = -1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '0'           AND ttext = tsqlt AND tbool = false AND tint2 =  0     AND tint4 =  0          AND tint8 =  0     AND treal =    0   AND tjson =           '0'::jsonb AND tdirn = 0"));
+    CHECK(1 == conn.get_count("nodes", "           tsqlt          = '0.5'         AND ttext = tsqlt AND tbool = true  AND tint2 =  0     AND tint4 =  0          AND tint8 =  0     AND treal =  0.5   AND tjson =         '0.5'::jsonb AND tdirn = 1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '1'           AND ttext = tsqlt AND tbool = true  AND tint2 =  1     AND tint4 =  1          AND tint8 =  1     AND treal =    1   AND tjson =           '1'::jsonb AND tdirn = 1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '2'           AND ttext = tsqlt AND tbool = true  AND tint2 =  2     AND tint4 =  2          AND tint8 =  2     AND treal =    2   AND tjson =           '2'::jsonb AND tdirn = 1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '32767'       AND ttext = tsqlt AND tbool = true  AND tint2 = 32767  AND tint4 = 32767       AND tint8 = 32767  AND treal = 32767  AND tjson =       '32767'::jsonb AND tdirn = 1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '32768'       AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 = 32768       AND tint8 = 32768  AND treal = 32768  AND tjson =       '32768'::jsonb AND tdirn = 1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '32769'       AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 = 32769       AND tint8 = 32769  AND treal = 32769  AND tjson =       '32769'::jsonb AND tdirn = 1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '2147483647'  AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 = 2147483647  AND tint8 = 2147483647                AND tjson =  '2147483647'::jsonb AND tdirn = 1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '2147483648'  AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 IS NULL       AND tint8 = 2147483648                AND tjson =  '2147483648'::jsonb AND tdirn = 1"));
+    CHECK(1 == conn.get_count("nodes", "split_part(tsqlt, '.', 1) = '2147483649'  AND ttext = tsqlt AND tbool = true  AND tint2 IS NULL  AND tint4 IS NULL       AND tint8 = 2147483649                AND tjson =  '2147483649'::jsonb AND tdirn = 1"));
     // clang-format on
 }
 
 TEST_CASE("type string (with bool)")
 {
-    testing::opt_t const options = testing::opt_t().flex(conf_file);
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
 
     REQUIRE_NOTHROW(
         db.run_import(options, "n10 v1 dV x10.0 y10.0 Ttype=string-bool\n"));
@@ -119,7 +127,7 @@ TEST_CASE("type string (with bool)")
 
 TEST_CASE("type string (with direction)")
 {
-    testing::opt_t const options = testing::opt_t().flex(conf_file);
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
 
     REQUIRE_NOTHROW(db.run_import(
         options, "n10 v1 dV x10.0 y10.0 Ttype=string-direction\n"));
@@ -133,7 +141,7 @@ TEST_CASE("type string (with direction)")
 
 TEST_CASE("type string (with number)")
 {
-    testing::opt_t const options = testing::opt_t().flex(conf_file);
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
 
     REQUIRE_NOTHROW(db.run_import(
         options, "n10 v1 dV x10.0 y10.0 Ttype=string-with-number\n"));
@@ -166,7 +174,7 @@ TEST_CASE("type string (with number)")
 
 TEST_CASE("type string (with invalid number)")
 {
-    testing::opt_t const options = testing::opt_t().flex(conf_file);
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
 
     REQUIRE_NOTHROW(db.run_import(
         options, "n10 v1 dV x10.0 y10.0 Ttype=string-with-invalid-number\n"));
@@ -188,7 +196,7 @@ TEST_CASE("type string (with invalid number)")
 
 TEST_CASE("type number in column where it doesn't belong")
 {
-    testing::opt_t const options = testing::opt_t().flex(conf_file);
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
 
     REQUIRE_THROWS(db.run_import(
         options, "n10 v1 dV x10.0 y10.0 Ttype=number-fail column=thstr\n"));
@@ -200,10 +208,11 @@ TEST_CASE("type number in column where it doesn't belong")
 
 TEST_CASE("Adding a function should always fail")
 {
-    testing::opt_t const options = testing::opt_t().flex(conf_file);
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
 
-    std::string const types[] = {"ttext", "tbool", "tint2", "tint4", "tint8",
-                                 "treal", "thstr", "tdirn", "tsqlt"};
+    std::array<std::string, 9> const types = {"ttext", "tbool", "tint2",
+                                              "tint4", "tint8", "treal",
+                                              "thstr", "tdirn", "tsqlt"};
 
     for (auto const &type : types) {
         auto const line =
@@ -218,7 +227,7 @@ TEST_CASE("Adding a function should always fail")
 
 TEST_CASE("type table")
 {
-    testing::opt_t const options = testing::opt_t().flex(conf_file);
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
 
     REQUIRE_NOTHROW(
         db.run_import(options, "n10 v1 dV x10.0 y10.0 Ttype=table\n"));
@@ -227,13 +236,15 @@ TEST_CASE("type table")
 
     CHECK(2 == conn.get_count("nodes"));
 
-    CHECK(1 == conn.get_count("nodes", "thstr = ''"));
-    CHECK(1 == conn.get_count("nodes", "thstr = 'a=>b,c=>d'"));
+    CHECK(1 == conn.get_count("nodes", "thstr = '' AND tjson = '{}'::jsonb"));
+    CHECK(1 == conn.get_count("nodes",
+                              "thstr = 'a=>b,c=>d' AND "
+                              "tjson = '{\"a\": \"b\", \"c\": \"d\"}'::jsonb"));
 }
 
 TEST_CASE("Adding a table with non-strings should fail for hstore")
 {
-    testing::opt_t const options = testing::opt_t().flex(conf_file);
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
 
     char const *const line = "n10 v1 dV x10.0 y10.0 Ttype=table-hstore-fail\n";
     REQUIRE_THROWS(db.run_import(options, line));
@@ -243,12 +254,12 @@ TEST_CASE("Adding a table with non-strings should fail for hstore")
     CHECK(0 == conn.get_count("nodes"));
 }
 
-TEST_CASE("Adding a table should fail except for hstore")
+TEST_CASE("Adding a table should fail except for hstore and json/jsonb")
 {
-    testing::opt_t const options = testing::opt_t().flex(conf_file);
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
 
-    std::string const types[] = {"ttext", "tbool", "tint2", "tint4",
-                                 "tint8", "treal", "tdirn", "tsqlt"};
+    std::array<std::string, 8> const types = {
+        "ttext", "tbool", "tint2", "tint4", "tint8", "treal", "tdirn", "tsqlt"};
 
     for (auto const &type : types) {
         auto const line =
@@ -260,3 +271,35 @@ TEST_CASE("Adding a table should fail except for hstore")
 
     CHECK(0 == conn.get_count("nodes"));
 }
+
+TEST_CASE("Adding a complex table in jsonb")
+{
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
+
+    char const *const line = "n10 v1 dV x10.0 y10.0 Ttype=json\n";
+    REQUIRE_NOTHROW(db.run_import(options, line));
+
+    auto conn = db.db().connect();
+
+    CHECK(1 ==
+          conn.get_count(
+              "nodes",
+              "tjson = '{"
+              "\"astring\": \"123\", \"aninteger\": 124, \"anumber\": 12.5, "
+              "\"atrue\": true, \"afalse\": false, "
+              "\"atable\": {\"a\": \"nested\", \"tab\": \"le\"}, "
+              "\"anarray\": [4, 3, 7]"
+              "}'::jsonb"));
+}
+
+TEST_CASE("Adding a table with a loop should fail")
+{
+    testing::opt_t const options = testing::opt_t().flex(CONF_FILE);
+
+    char const *const line = "n10 v1 dV x10.0 y10.0 Ttype=json-loop\n";
+    REQUIRE_THROWS(db.run_import(options, line));
+
+    auto conn = db.db().connect();
+
+    CHECK(0 == conn.get_count("nodes"));
+}
